@@ -1,33 +1,18 @@
+import copy
 import json
-import jwt
+import os
 import time
 import uuid
-import copy
-import os
 
-from pydal.objects import FieldVirtual
-
-from py4web import request, response, HTTP
-from py4web.utils.param import Param
+import jwt
 from pydal._compat import to_native
+from pydal.objects import FieldVirtual
 from pydal.validators import Validator
+from yatl.helpers import (DIV, FORM, INPUT, LABEL, OPTION, SELECT, SPAN, TABLE,
+                          TD, TEXTAREA, TR, XML, A, P)
 
-from yatl.helpers import (
-    A,
-    TEXTAREA,
-    INPUT,
-    TR,
-    TD,
-    TABLE,
-    DIV,
-    LABEL,
-    FORM,
-    SELECT,
-    OPTION,
-    P,
-    SPAN,
-    XML,
-)
+from py4web import HTTP, request, response
+from py4web.utils.param import Param
 
 
 def to_id(field):
@@ -87,7 +72,7 @@ class DateTimeWidget:
 
     def make(self, field, value, error, title, placeholder="", readonly=False):
         return INPUT(
-            _value=str(value).replace(" ", "T"),
+            _value=str(value).replace(" ", "T")[:16],
             _type=self.input_type,
             _id=to_id(field),
             _name=field.name,
@@ -292,6 +277,9 @@ class FormStyleFactory:
         form_action = request.url.split(":", 1)[1]
         form_enctype = "multipart/form-data"
 
+        if "_method" in kwargs:
+            form_method = kwargs.pop("_method")
+
         form = FORM(
             _method=form_method, _action=form_action, _enctype=form_enctype, **kwargs
         )
@@ -604,7 +592,7 @@ FormStyleBulma.classes.update(
         "inner": "control",
         "label": "label",
         "info": "help",
-        "error": "help is-danger py4web-validation-error",
+        "error": "help is-danger py4web-validation-error mt-1",
         "submit": "button is-primary",
         "input": "input",
         "input[type=text]": "input",
@@ -754,8 +742,13 @@ class Form(object):
 
         if self.record:
             self.vars = self._read_vars_from_record(table)
-        if not readonly and request.method != "GET":
-            post_vars = request.POST
+
+        if not readonly:  # and request.method != "GET":
+            if request.method == "GET":
+                post_vars = request.GET
+            else:
+                post_vars = request.POST
+
             form_vars = copy.deepcopy(request.forms)
             for k in form_vars:
                 self.vars[k] = form_vars[k]
@@ -764,9 +757,10 @@ class Form(object):
 
             # We only a process a form if it is POST and the formkey matches (correct formname and crsf)
             # Notice: we never expose the crsf uuid, we only use to sign the form uuid
-            if request.method == "POST":
+            if request.method in ["GET", "POST"]:
                 if not self.csrf_protection or self._verify_form(post_vars):
                     process = True
+
             if process:
                 record_id = self.record and self.record.get("id")
                 if not post_vars.get("_delete"):
@@ -794,7 +788,6 @@ class Form(object):
                                 continue
                             if field.type == "upload":
                                 value = request.files.get(field.name)
-                                print(str(value)[:100])
                                 delete = post_vars.get("_delete_" + field.name)
                                 if value is not None:
                                     if field.uploadfolder:
@@ -924,11 +917,6 @@ class Form(object):
 
                 button_attributes = item.attributes
                 button_attributes["_label"] = item.children[0]
-                button_attributes["_type"] = (
-                    button_attributes.pop("_role")
-                    if "_role" in button_attributes
-                    else None
-                )
                 helper["json_controls"]["form_buttons"] += [button_attributes]
 
             if self.action:
