@@ -7,28 +7,29 @@ import copy
 import datetime
 from urllib.parse import urlparse
 
-from pydal.objects import Field, FieldVirtual
+import ombott
+from pydal.objects import Field, FieldVirtual, Expression
 from yatl.helpers import (
     CAT,
     DIV,
+    FORM,
+    INPUT,
+    OPTION,
+    SELECT,
+    SPAN,
     TABLE,
+    TAG,
     TBODY,
-    TR,
     TD,
     TH,
-    A,
-    SPAN,
-    I,
     THEAD,
-    TAG,
-    INPUT,
+    TR,
     XML,
-    FORM,
-    SELECT,
-    OPTION,
+    A,
+    I,
 )
 
-from py4web import request, URL, redirect, HTTP
+from py4web import HTTP, URL, redirect, request
 from py4web.utils.form import Form, FormStyleDefault, join_classes
 from py4web.utils.param import Param
 
@@ -352,6 +353,7 @@ class Grid:
         formstyle=FormStyleDefault,
         grid_class_style=GridClassStyle,
         T=lambda text: text,
+        groupby=None,
         # deprecated
         fields=None,
     ):
@@ -395,6 +397,7 @@ class Grid:
             show_id=show_id,
             orderby=orderby,
             left=left,
+            groupby=groupby,
             search_form=search_form,
             search_queries=search_queries,
             headings=headings or [],
@@ -446,7 +449,16 @@ class Grid:
         if auto_process:
             self.process()
 
+    def is_creatable(self):
+        if self.param.groupby:
+            return False
+        if callable(self.param.create):
+            return self.param.create()
+        return self.param.create
+
     def is_editable(self, row):
+        if self.param.groupby:
+            return False
         if callable(self.param.editable):
             return self.param.editable(row)
         return self.param.editable
@@ -457,6 +469,8 @@ class Grid:
         return self.param.details
 
     def is_deletable(self, row):
+        if self.param.groupby:
+            return False
         if callable(self.param.deletable):
             return self.param.deletable(row)
         return self.param.deletable
@@ -481,6 +495,7 @@ class Grid:
 
         parts = self.path.split("/")
         self.action = parts[0] or "select"
+
         if self.param.field_id:
             self.tablename = str(self.param.field_id._table)
         else:
@@ -493,6 +508,17 @@ class Grid:
         if not self.param.columns:
             # if no column specified use all fields
             self.param.columns = [field for field in table if field.readable]
+
+        #  if any columns are Expression but not Field, get the field info from 'first' attribute
+        converted_columns = []
+        for col in self.param.columns:
+            if isinstance(col, Expression) and not isinstance(col, Field):
+                converted_columns.append(col.first)
+            else:
+                converted_columns.append(col)
+        self.param.columns = converted_columns
+
+        if not self.param.columns:
             self.needed_fields = self.param.columns[:]
         elif any(isinstance(col, Column) for col in self.param.columns):
             # if we use columns we have to get all fields and assume a single table
@@ -540,6 +566,28 @@ class Grid:
         else:
             record = None
 
+        #  ensure the user has access for new/details/edit/delete if chosen
+        if self.action == "new" and not self.is_creatable():
+            raise HTTP(
+                403,
+                f"You do not have access to create a record in the {self.tablename} table.",
+            )
+        if self.action == "details" and not self.is_readable(record):
+            raise HTTP(
+                403,
+                f"You do not have access to read a record from the {self.tablename} table.",
+            )
+        if self.action == "edit" and not self.is_editable(record):
+            raise HTTP(
+                403,
+                f"You do not have access to edit a record in the {self.tablename} table.",
+            )
+        if self.action == "delete" and not self.is_deletable(record):
+            raise HTTP(
+                403,
+                f"You do not have access to delete a record in the {self.tablename} table.",
+            )
+
         if self.action in ["new", "details", "edit"]:
 
             readonly = self.action == "details"
@@ -560,7 +608,7 @@ class Grid:
                     self.form.param.sidecar.append(self.param.new_sidecar)
                 if self.param.new_submit_value:
                     self.form.param.submit_value = self.param.new_submit_value
-            if self.action == "details":
+            if self.action == "details" and self.is_readable(record):
                 if self.param.details_sidecar:
                     self.form.param.sidecar.append(self.param.details_sidecar)
                 if self.param.details_submit_value:
@@ -625,7 +673,15 @@ class Grid:
             if self.param.left:
                 select_params["left"] = self.param.left
 
-            if self.param.left:
+            if self.param.groupby:
+                select_params["groupby"] = self.param.groupby
+
+            if self.param.groupby:
+                #  need groupby fields in select to get proper count
+                self.total_number_of_rows = len(
+                    db(query).select(*self.param.groupby, **select_params)
+                )
+            elif self.param.left:
                 # TODO: maybe this can be made more efficient
                 self.total_number_of_rows = len(
                     db(query).select(db[self.tablename].id, **select_params)
@@ -1048,6 +1104,12 @@ class Grid:
                         # if None was returned, no button is available for this row: ignore this value in the
                         # list
                         continue
+                attrs = (
+                    self.attributes_plugin.confirm(message=self.T(btn.message))
+                    if btn.message and btn.message != ""
+                    else dict()
+                )
+
                 cat.append(
                     self._make_action_button(
                         url=btn.url,
@@ -1060,6 +1122,7 @@ class Grid:
                         ignore_attribute_plugin=btn.ignore_attribute_plugin
                         if "ignore_attribute_plugin" in btn.__dict__
                         else False,
+                        **attrs,
                     )
                 )
 
@@ -1124,6 +1187,11 @@ class Grid:
                         # if None was returned, no button is available for this row: ignore this value in the
                         # list
                         continue
+                attrs = (
+                    self.attributes_plugin.confirm(message=self.T(btn.message))
+                    if btn.message and btn.message != ""
+                    else dict()
+                )
                 cat.append(
                     self._make_action_button(
                         url=btn.url,
@@ -1136,6 +1204,7 @@ class Grid:
                         ignore_attribute_plugin=btn.ignore_attribute_plugin
                         if "ignore_attribute_plugin" in btn.__dict__
                         else False,
+                        **attrs,
                     )
                 )
 
