@@ -1,9 +1,10 @@
-import json
+import base64
 import hashlib
+import json
 import time
 import uuid
-import base64
-from py4web import request, abort
+
+from py4web import HTTP, request
 from py4web.core import Fixture, Session
 
 
@@ -20,16 +21,16 @@ class URLVerifier(Fixture):
             self.__prerequisites__ = [url_signer.session]
         self.url_signer = url_signer
 
-    def on_request(self):
+    def on_request(self, context):
         """Checks the request's signature"""
         # extra and remove the signature from the query
         signature = request.query.get("_signature")
         if signature is None:
-            abort(403)
-        try:
+            raise HTTP(403)
+        if 1: # try:
             h = self.url_signer.algo(self.url_signer.get_key())
             signature = request.query["_signature"]
-            sig_content = base64.b85decode(signature.encode("utf-8")).decode("utf-8")
+            sig_content = base64.b16decode(signature.encode("utf-8")).decode("utf-8")
             sig_dict = json.loads(sig_content)
             ts = sig_dict["ts"]
             salt = sig_dict["salt"]
@@ -39,21 +40,21 @@ class URLVerifier(Fixture):
                     request.fullpath, request.query, ts, salt
                 )
             )
-            computed_sig = base64.b85encode(h.digest()).decode("utf-8")
+            computed_sig = base64.b16encode(h.digest()).decode("utf-8")
             if sig != computed_sig:
-                abort(403)
+                raise HTTP(403)
             # We remove the signature, not to pollute the request.
             del request.query["_signature"]
             # Checks the expiration time.
             if self.url_signer.lifespan is not None:
                 if float(ts) + self.url_signer.lifespan < time.time():
-                    abort(403)
-        except:
-            abort(403)
+                    raise HTTP(403)
+        #except:
+        #    raise HTTP(403)
 
     def _decode_ts(self, ts_string):
         """Decodes the timestamp, removing the salt."""
-        s = base64.b85encode(ts_string.encode("utf-8")).decode("utf-8")
+        s = base64.b16encode(ts_string.encode("utf-8")).decode("utf-8")
         return float(s.split(";")[1])
 
 
@@ -114,16 +115,28 @@ class URLSigner(Fixture):
         assert "_signature" not in self.variables_to_sign
         self.algo = algo or hashlib.sha256
 
+    #@property
+    #def local(self):
+    #    return self._safe_local        
+
+    def on_request(self, context):
+        """Creates the signing key if necessary."""
+        print("on_request", self.session.get("_signature_key")) 
+        if self.session is not None and self.session.get("_signature_key") is None:
+            key = str(uuid.uuid1())
+            self.session["_signature_key"] = key
+            # Note that we CANNOT save the key in the URLsigner object,
+            # because there is only one object -- it's not thread-local in
+            # the way the session is.  So we cannot cache it for the later
+            # call to get_key.
+
     def get_key(self):
-        """Gets the signing key, creating it if necessary."""
+        """Gets the signing key."""
         if self.session is None:
             key = self.key
-            assert self.key is not None, "You need to specify a signing key"
         else:
             key = self.session.get("_signature_key")
-            if key is None:
-                key = str(uuid.uuid1())
-                self.session["_signature_key"] = key
+        assert key is not None, "No signing key"
         return key.encode("utf8")
 
     def get_info_to_sign(self, url, variables, ts, salt):
@@ -147,9 +160,9 @@ class URLSigner(Fixture):
         ts = "%.3f" % time.time()
         salt = str(uuid.uuid1())
         h.update(self.get_info_to_sign(url, variables, ts, salt))
-        sig = base64.b85encode(h.digest()).decode("utf-8")
+        sig = base64.b16encode(h.digest()).decode("utf-8")
         sig_content = json.dumps(dict(ts=ts, salt=salt, sig=sig))
-        signature = base64.b85encode(sig_content.encode("utf-8")).decode("utf-8")
+        signature = base64.b16encode(sig_content.encode("utf-8")).decode("utf-8")
         variables["_signature"] = signature
 
     def verify(self):
